// Copyright (C) 2023 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only

#ifndef QTQUICK3DRUNTIMERENDERGLOBAL_H
#define QTQUICK3DRUNTIMERENDERGLOBAL_H

//
//  W A R N I N G
//  -------------
//
// This file is part of the QtQuick3D API, with limited compatibility guarantees.
// Usage of this API may make your code source and binary incompatible with
// future versions of Qt.
//

#include <QtQuick3DRuntimeRender/qtquick3druntimerenderexports.h>

QT_BEGIN_NAMESPACE

using QSSGPassKey = void *;

QT_END_NAMESPACE

#endif // QTQUICK3DRUNTIMERENDERGLOBAL_H
