// Copyright (C) 2019 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only

#ifndef QTQUICK3DASSETIMPORTGLOBAL_H
#define QTQUICK3DASSETIMPORTGLOBAL_H

//
//  W A R N I N G
//  -------------
//
// This file is not part of the Qt API.  It exists purely as an
// implementation detail.  This header file may change from version to
// version without notice, or even be removed.
//
// We mean it.
//

#include <QtGui/qtguiglobal.h>
#include <QtQuick3DAssetImport/qtquick3dassetimportexports.h>

#endif // QTQUICK3DASSETIMPORTGLOBAL_H
