// Copyright (C) 2021 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only

#ifndef QTQUICK3DPARTICLESGLOBAL_H
#define QTQUICK3DPARTICLESGLOBAL_H

#include <QtCore/qglobal.h>
#include <QtQuick3DParticles/qtquick3dparticlesexports.h>

#endif // QTQUICK3DPARTICLESGLOBAL_H
